# coding:utf-8

from .formatting import qformat
from .restructables import RestructableList as RL, RestructableDict as RD


class Client(object):
    def __init__(self, dbapi, **kwargs):
        self._dbapi = dbapi
        self._config = kwargs

        self._connection = None
        self._cursor = None

        self._result_cols = None
        self._result_rows = None
        self._lastrowid = None
        self._rowcount = None

    def connection(self):
        if self._connection is None:
            self._connection = self._dbapi.connect(**self._config)
        return self._connection

    def cursor(self):
        if self._cursor is None:
            self._cursor = self.connection().cursor()
        return self._cursor

    def commit(self, close=True):
        self.connection().commit()

        if close:
            self.close()

        return self

    def rollback(self, close=True):
        self.connection().rollback()

        if close:
            self.close()

        return self

    def close(self):
        if self._cursor is not None:
            self._cursor.close()
            self._cursor = None

        if self._connection is not None:
            self._connection.close()
            self._connection = None

        return self

    def exes(self, fmt, **kwargs):
        return self.exe(True, fmt, **kwargs)

    def exew(self, fmt, **kwargs):
        return self.exe(False, fmt, **kwargs)

    def exe(self, strong, fmt, **kwargs):
        try:
            self.cursor().execute(qformat(fmt, **kwargs))

        except Exception as e:
            self.close()
            raise e

        else:
            self._lastrowid = self.cursor().lastrowid
            self._rowcount = self.cursor().rowcount

            self._result_cols = [c[0] for c in self.cursor().description or []]
            self._result_rows = [r for r in self.cursor()]

            if not strong:
                self.close()

            return self

    def scalar(self):
        if len(self._result_rows) > 0:
            return self._result_rows[0][0]

        else:
            return None

    def lastid(self):
        return self._lastrowid

    def rowcount(self):
        return self._rowcount

    def first(self):
        if len(self._result_rows) > 0:
            return RD(dict(zip(self._result_cols, self._result_rows[0])))

        else:
            return None

    def all(self):
        return RL([dict(zip(self._result_cols, r)) for r in self._result_rows])



# coding:utf-8

import re
from datetime import datetime

PLACEHOLDER = re.compile(r'.:+[0-9a-zA-Z_]+')


def qformat(fmt, **kwargs):
    def _(m):
        g = m.group(0)
        pre, key = g[0], g[1:]

        if pre == '\\':
            return key

        else:
            v = kwargs[key.lstrip(':')]
            v = str(v) if key.startswith('::') else escape(v)
            return pre + v

    return PLACEHOLDER.sub(_, fmt)


def escape(v):
    if isinstance(v, (int, long, float)):
        return str(v)

    elif isinstance(v, bool):
        return '1' if v else '0'

    elif isinstance(v, (str, unicode)):
        return "'{}'".format(str(v).replace("'", "\\'"))

    elif v is None:
        return 'NULL'

    elif isinstance(v, (list, tuple, set)):
        return ','.join([value(e) for e in v])

    elif isinstance(v, datetime):
        return "'{}'".format(v.strftime('%Y-%m-%d %H:%M:%S'))

    raise TypeError('Incompatible type :{}'.format(v))



# coding:utf-8


def _mklist(key, separator):
    def _(e):
        e[key] = e[key].split(separator)
    return _


def _mkdict(**rules):
    def _(e):
        for fromkey, tokey in rules.items():
            keypath = tokey.split('.')
            head = e
            for k in keypath[:-1]:
                head = head.setdefault(k, {})
            head[keypath[-1]] = e[fromkey]
            del e[fromkey]
    return _


class RestructableList(list):
    def process(self, f):
        for e in self:
            f(e)
        return self

    def mklist(self, key, separator=','):
        return self.process(_mklist(key, separator))

    def mkdict(self, **rules):
        return self.process(_mkdict(**rules))


class RestructableDict(dict):
    def process(self, f):
        f(self)
        return self

    def mklist(self, key, separator=','):
        return self.process(_mklist(key, separator))

    def mkdict(self, **rules):
        return self.process(_mkdict(**rules))



# coding:utf-8

import os


class Storage(dict):
    def load(self, path):
        for f in self._load(path):
            with open(f, 'r') as fp:
                key = f[len(path):].lstrip('/')[:-4]
                self[key] = self._make_compact(fp.read())

    def _load(self, path):
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith('.sql'):
                    yield os.path.join(root, f)

    def _make_compact(self, q):
        all_lines = q.splitlines()
        query_lines = [l for l in all_lines if not l.startswith(('#', '--'))]
        query = ' '.join(query_lines)
        return ' '.join(query.split())
